//--------------------------------------------------------------
// File     : stm32_ub_am2301.c
// Datum    : 20.12.2013
// Version  : 1.0
// Autor    : UB
// EMail    : mc-4u(@)t-online.de
// Web      : www.mikrocontroller-4u.de
// CPU      : STM32F4
// IDE      : CooCox CoIDE 1.7.4
// GCC      : 4.7 2012q4
// Module   : GPIO, TIM, MISC 
// Funktion : Temperatur- und Feuchtesensor (Aosong AM2301)
//
// Data-Pin : PD3 (ein externer PullUp von 4k7 ist notwendig)
// Hinweis  : benutzt wird Timer-7
//            (Protokoll ist nicht kompatibel mit "One-Wire")
//--------------------------------------------------------------


//--------------------------------------------------------------
// Includes
//--------------------------------------------------------------
#include "stm32_ub_am2301.h"


//--------------------------------------------------------------
// globale Variabeln
//--------------------------------------------------------------
uint32_t tim7_akt_delay;         // fuer den Timer


//--------------------------------------------------------------
// interne Funktionen
//--------------------------------------------------------------
void P_AM2301_InitGPIO(void);
void P_AM2301_DataLo(void);
void P_AM2301_DataHi(void);
BitAction P_AM2301_DataRead(void);
void P_AM2301_InitTIM(void);
void P_AM2301_InitNVIC(void);
void P_AM2301_Delay_us(uint32_t delay);
uint32_t P_AM2301_Mess_Lo_us(uint32_t delay);
uint32_t P_AM2301_Mess_Hi_us(uint32_t delay);
uint32_t P_AM2301_Wait_Lo_us(uint32_t delay);
ErrorStatus P_AM2301_ReadSequenz(void);
uint8_t P_AM2301_ReadByte(void);
BitAction P_AM2301_ReadBit(void);


//--------------------------------------------------------------
// Init-Funktion vom AM2301
//--------------------------------------------------------------
void UB_AM2301_Init(void)
{
  // init der Variabeln
  AM2301.humidity=0.0;
  AM2301.temperature=0.0;
  tim7_akt_delay=0;

  // GPIO init
  P_AM2301_InitGPIO();

  // Timer init
  P_AM2301_InitTIM();

  // NVIC init
  P_AM2301_InitNVIC();

  // Timer auf 0 setzen
  tim7_akt_delay=0;
  // Timer starten
  TIM_Cmd(TIM7, ENABLE);
}


//--------------------------------------------------------------
// Auslesen vom AM2301
// (das auslesen kann nur alle 2sec gemacht werden)
//
// Return_wert :
//   AM2301_NOTREADY : Sensor noch nicht bereit
//   AM2301_ERROR    : fehler beim auslesen
//   AM2301_OK       : auslesen ok -> Daten stehen in "AM2301" 
//
// AM2301.humidity    :   [0.0 bis 99.9] Luftfeuchte in Prozent
// AM2301.temperature : [-40.0 bis 80.0] Temperatur in Grad Celsius
//--------------------------------------------------------------
AM2301_STATUS_t UB_AM2301_Read(void)
{
  AM2301_STATUS_t ret_wert=AM2301_NOTREADY;

  // check ob 2sec abgelaufen
  if(tim7_akt_delay>2000000) {      
    // Timer disable
    TIM_Cmd(TIM7, DISABLE);
    // Sensor auslesen
    if(P_AM2301_ReadSequenz()==SUCCESS) {
      ret_wert=AM2301_OK;
    }
    else {
      ret_wert=AM2301_ERROR;
    }

    // Timer auf 0 setzen
    tim7_akt_delay=0;
    // Timer starten
    TIM_Cmd(TIM7, ENABLE);
  } 

  return(ret_wert);  
}


//--------------------------------------------------------------
// interne Funktion
// init vom GPIO-Pin
//--------------------------------------------------------------
void P_AM2301_InitGPIO(void)
{
  GPIO_InitTypeDef  GPIO_InitStructure;

  // Clock enable
  RCC_AHB1PeriphClockCmd(AM2301_CLK, ENABLE);

  // Config als OpenDrain-Ausgang
  GPIO_InitStructure.GPIO_Pin = AM2301_PIN;
  GPIO_InitStructure.GPIO_Mode = GPIO_Mode_OUT;
  GPIO_InitStructure.GPIO_OType = GPIO_OType_OD;
  GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_UP;
  GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
  GPIO_Init(AM2301_PORT, &GPIO_InitStructure);

  // Hi-Pegel ausgeben
  P_AM2301_DataHi();
}


//--------------------------------------------------------------
// interne Funktion
// Daten-Pin auf Lo-Pegel
//--------------------------------------------------------------
void P_AM2301_DataLo(void)
{
  AM2301_PORT->BSRRH = AM2301_PIN;
}


//--------------------------------------------------------------
// interne Funktion
// Daten-Pin auf Tri-State
//--------------------------------------------------------------
void P_AM2301_DataHi(void)
{
  AM2301_PORT->BSRRL = AM2301_PIN;
}


//--------------------------------------------------------------
// interne Funktion
// Pegel am Daten-Pin auslesen
//  -> bei Lo-Pegel = Bit_RESET
//  -> bei Hi-Pegel = Bit_SET
//--------------------------------------------------------------
BitAction P_AM2301_DataRead(void)
{
  BitAction ret_wert;

  ret_wert=GPIO_ReadInputDataBit(AM2301_PORT, AM2301_PIN);

  return(ret_wert);
}


//--------------------------------------------------------------
// interne Funktion
// init vom Timer
//--------------------------------------------------------------
void P_AM2301_InitTIM(void)
{
  TIM_TimeBaseInitTypeDef  TIM_TimeBaseStructure;

  // Clock enable
  RCC_APB1PeriphClockCmd(RCC_APB1Periph_TIM7, ENABLE);

  // Timer disable
  TIM_Cmd(TIM7, DISABLE);

  // Timer auf 1us einstellen
  TIM_TimeBaseStructure.TIM_Period=AM2301_PERIOD;
  TIM_TimeBaseStructure.TIM_Prescaler=AM2301_PRESCALE;
  TIM_TimeBaseStructure.TIM_ClockDivision=TIM_CKD_DIV1;
  TIM_TimeBaseStructure.TIM_CounterMode=TIM_CounterMode_Up;
  TIM_TimeBaseInit(TIM7, &TIM_TimeBaseStructure);

  TIM_ARRPreloadConfig(TIM7, ENABLE);
}


//--------------------------------------------------------------
// interne Funktion
// init vom NVIC
//--------------------------------------------------------------
void P_AM2301_InitNVIC(void)
{
  NVIC_InitTypeDef NVIC_InitStructure;

  // Update Interrupt enable
  TIM_ITConfig(TIM7,TIM_IT_Update,ENABLE);

  // NVIC konfig
  NVIC_InitStructure.NVIC_IRQChannel = TIM7_IRQn;
  NVIC_InitStructure.NVIC_IRQChannelPreemptionPriority = 0;
  NVIC_InitStructure.NVIC_IRQChannelSubPriority = 0;
  NVIC_InitStructure.NVIC_IRQChannelCmd = ENABLE;
  NVIC_Init(&NVIC_InitStructure);
}


//------------------------------------------------------------
// interne Funktion
// Delay-Funktion (us)
//------------------------------------------------------------
void P_AM2301_Delay_us(uint32_t delay)
{
  // Timer auf 0 setzen
  tim7_akt_delay=0;
  // Timer starten
  TIM_Cmd(TIM7, ENABLE);
  // warten bis Timer abgelaufen
  while(tim7_akt_delay<delay);
  // Timer stoppen
  TIM_Cmd(TIM7, DISABLE);
}


//------------------------------------------------------------
// interne Funktion
// Mess-Funktion (us)
// Laenge vom Lo-Impuls messen
//------------------------------------------------------------
uint32_t P_AM2301_Mess_Lo_us(uint32_t delay)
{
  uint32_t ret_wert=0;
  uint32_t start_pos=0, stop_pos=0;

  // Timer auf 0 setzen
  tim7_akt_delay=0;
  // Timer starten
  TIM_Cmd(TIM7, ENABLE);
  // warten bis Timer abgelaufen
  do {
    // Pegel am Data-Pin messen
    if(P_AM2301_DataRead()==Bit_RESET) {
      if(start_pos==0) start_pos=tim7_akt_delay;
    }
    else {
      if(start_pos>0) {
        if(stop_pos==0) stop_pos=tim7_akt_delay;
      }
    }
  }while((tim7_akt_delay<delay) && (stop_pos==0));
  // Timer stoppen
  TIM_Cmd(TIM7, DISABLE);

  // Lo-Impuls ausrechnen
  if(stop_pos>start_pos) {
    ret_wert=stop_pos-start_pos;
  }

  return(ret_wert);
}


//------------------------------------------------------------
// interne Funktion
// Mess-Funktion (us)
// Laenge vom Hi-Impuls messen
//------------------------------------------------------------
uint32_t P_AM2301_Mess_Hi_us(uint32_t delay)
{
  uint32_t ret_wert=0;
  uint32_t start_pos=0, stop_pos=0;

  // Timer auf 0 setzen
  tim7_akt_delay=0;
  // Timer starten
  TIM_Cmd(TIM7, ENABLE);
  // warten bis Timer abgelaufen
  do {
    // Pegel am Data-Pin messen
    if(P_AM2301_DataRead()==Bit_SET) {
      if(start_pos==0) start_pos=tim7_akt_delay;
    }
    else {
      if(start_pos>0) {
        if(stop_pos==0) stop_pos=tim7_akt_delay;
      }
    }
  }while((tim7_akt_delay<delay) && (stop_pos==0));
  // Timer stoppen
  TIM_Cmd(TIM7, DISABLE);

  // Hi-Impuls ausrechnen
  if(stop_pos>start_pos) {
    ret_wert=stop_pos-start_pos;
  }

  return(ret_wert);
}


//------------------------------------------------------------
// interne Funktion
// wartet auf Lo-Signal (oder Timeout)
//------------------------------------------------------------
uint32_t P_AM2301_Wait_Lo_us(uint32_t delay)
{
  uint32_t ret_wert=0;
  uint32_t stop=0;

  // Timer auf 0 setzen
  tim7_akt_delay=0;
  // Timer starten
  TIM_Cmd(TIM7, ENABLE);
  // warten bis Timer abgelaufen
  do {
    // Pegel am Data-Pin messen
    if(P_AM2301_DataRead()==Bit_RESET) {
      stop=1;
      ret_wert=1;
    }
  }while((tim7_akt_delay<delay) && (stop==0));
  // Timer stoppen
  TIM_Cmd(TIM7, DISABLE);

  return(ret_wert);
}


//--------------------------------------------------------------
// interne Funktion
// auslesen der Sensor-Daten (5 x 8 Bit lesen)
// Return_wert :
//  -> ERROR   , wenn kein Response-Signal empfangen wurde
//  -> SUCCESS , wenn OK 
//--------------------------------------------------------------
ErrorStatus P_AM2301_ReadSequenz(void)
{
  ErrorStatus ret_wert=ERROR; 
  uint32_t lo_time=0;
  uint8_t h_hi,h_lo;
  uint8_t t_hi,t_lo;
  uint8_t parity,check;
  uint16_t dummy;

  //-------------------------------
  // start sequenz
  //-------------------------------
  P_AM2301_DataLo();
  P_AM2301_Delay_us(1000); // 1ms  
  P_AM2301_DataHi();
  //-------------------------------
  // messen vom response signal
  //-------------------------------
  lo_time=P_AM2301_Mess_Lo_us(200); // 75<T<85 (ist=90us)
  if((lo_time>=60) && (lo_time<=100)) {
    //-------------------------------
    // warte auf erste Lo-Flanke
    //-------------------------------
    if(P_AM2301_Wait_Lo_us(150)==0) return(ERROR); // keine Lo-Flanke
    //-------------------------------
    // 5 Datenbytes lesen
    //-------------------------------
    h_hi=P_AM2301_ReadByte();
    h_lo=P_AM2301_ReadByte();
    t_hi=P_AM2301_ReadByte();
    t_lo=P_AM2301_ReadByte();
    parity=P_AM2301_ReadByte();

    //-------------------------------
    // Daten auswerten
    //-------------------------------
    check=(h_hi+h_lo+t_hi+t_lo);
    if(check==parity) {
      // parity ok
      ret_wert=SUCCESS;
      // humidity ausrechnen 
      dummy=(h_hi<<8)|h_lo;
      AM2301.humidity=(float)(dummy/10.0);
      // temperatur ausrechnen
      dummy=(t_hi<<8)|t_lo;
      if((dummy&0x8000)==0) {
        // positiv
        AM2301.temperature=(float)(dummy/10.0);
      }
      else {
        // negativ
        dummy&=0x7FFF;
        AM2301.temperature=(float)(dummy/10.0);
        AM2301.temperature=0.0 - AM2301.temperature;
      }
    }
    else {
      // parity fehler
      ret_wert=ERROR;
    }    
  }
  else {
    // kein response signal
    ret_wert=ERROR;
  }

  return(ret_wert);
}


//------------------------------------------------------------
// interne Funktion
// auslesen von einem Byte (MSB first)
//------------------------------------------------------------
uint8_t P_AM2301_ReadByte(void)
{
  uint8_t ret_wert=0;
  uint8_t bit_maske=0x80,n;

  // 8bit lesen
  for(n=0;n<8;n++) {
    if(P_AM2301_ReadBit()==Bit_SET) {
      ret_wert|=bit_maske;
    }
    bit_maske=bit_maske>>1;
  }

  return(ret_wert);
}


//--------------------------------------------------------------
// interne Funktion
// auslesen von einem Bit
// (je nach laenge vom Hi-Pegel)
//  -> bei Lo = Bit_RESET
//  -> bei Hi = Bit_SET
//--------------------------------------------------------------
BitAction P_AM2301_ReadBit(void)
{
  BitAction ret_wert;
  uint32_t hi_time=0;

  // messen der Hi-Zeit
  // Lo => 22<T<30  (ist=25)
  // Hi => 68<T<75  (ist=76)
  hi_time=P_AM2301_Mess_Hi_us(150);
  if(hi_time>49) {
    ret_wert=Bit_SET;
  } 
  else {
    ret_wert=Bit_RESET;
  }  

  return(ret_wert);
}


//------------------------------------------------------------
// Timer-7 ISR
// wird alle 1us aufgerufen
//------------------------------------------------------------
void TIM7_IRQHandler(void)
{
  if (TIM_GetITStatus(TIM7, TIM_IT_Update) != RESET) {
    // wenn Interrupt aufgetreten
    TIM_ClearITPendingBit(TIM7, TIM_IT_Update);
 
    tim7_akt_delay++;
  }
}
